package reproducer;

import static org.junit.jupiter.api.Assertions.assertFalse;

import java.util.ArrayList;
import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;

import javax.validation.ConstraintViolation;
import javax.validation.ElementKind;
import javax.validation.Path;
import javax.validation.metadata.ConstraintDescriptor;
import javax.ws.rs.core.MediaType;

import org.jboss.resteasy.api.validation.ResteasyConstraintViolation;
import org.jboss.resteasy.api.validation.ResteasyViolationException;
import org.jboss.resteasy.api.validation.SimpleViolationsContainer;
import org.jboss.resteasy.plugins.providers.validation.ViolationsContainer;
import org.junit.jupiter.api.Test;

/**
 * This Test simulates the steps leading to the ConcurrentModificationException by creating the exception and bruteforcing
 * concurrent calls of the ResteasyViolationException#toString method.
 *
 * This is simpler than bruteforcing the ConcurrentModificationException with REST calls.
 *
 * When validation annotations are used directly in jax-rs methods, the resulting org.jboss.resteasy.api.validation.ResteasyViolationException
 * is created by resteasy itself.
 * Debugging showed, that the Exception is created with the constructor org.jboss.resteasy.api.validation.ResteasyViolationException#ResteasyViolationException(org.jboss.resteasy.api.validation.SimpleViolationsContainer,
 * java.util.List)
 *
 * This is simulated in the test execBruteforce()
 *
 * The other tests show that reason for the Exception is, that ResteasyViolationException#toString changes the state of the
 * exception in a not threadsafe way.
 * But the state changes only if ResteasyViolationException#violationLists isn't built in the constructor.
 */
public class ResteasyViolationExceptionToStringReproducerTests {

    @Test
    void testUnitary() {
        Set<ConstraintViolation<Object>> violations = new HashSet<>();
        violations.add(new DummyConstraintViolation());
        violations.add(new DummyConstraintViolation());
        violations.add(new DummyConstraintViolation());
        final SimpleViolationsContainer container = new SimpleViolationsContainer(violations);
        final List<MediaType> accept = new ArrayList<>();
        ResteasyViolationException violationException = new ResteasyViolationException(container, accept);
        assertFalse(violationException.toString().isEmpty());
    }

    /**
     * Create a ThreadPool with 10 Threads and 10 Runners who constantly call toString on the ResteasyViolationException.
     * In main-thread create multiple new exceptions for the runners to use.
     *
     * AtomicReference used to simplify the use of a single shared exception instance.
     *
     *
     * Scenario is scaled up to reduce the time until an exception occurs., but exceptions occurred in production.
     *
     * UseCase: Interceptor that logs the exception with an async-Logger and BusinessCode that converts the stacktrace-elements into
     * a json-response.
     *
     * Possible solution: call ResteasyViolationException#convertToStrings in all constructors.
     */
    @Test
    void execBruteforce() throws Exception {
        //setup runners
        ExecutorService executor = Executors.newFixedThreadPool(10);
        AtomicBoolean run = new AtomicBoolean(true);
        AtomicReference<ResteasyViolationException> exceptionToUse = new AtomicReference<>();

        //start runners
        List<Future<Boolean>> runnerResults = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            runnerResults.add(executor.submit(new ExceptionToStringCaller(run, exceptionToUse)));
        }

        //start create exceptions
        for (int i = 0; i < 1_000; i++) {
            Set<ConstraintViolation<Object>> violations = new HashSet<>();
            violations.add(new DummyConstraintViolation());
            violations.add(new DummyConstraintViolation());
            violations.add(new DummyConstraintViolation());
            final SimpleViolationsContainer container = new SimpleViolationsContainer(violations);
            final List<MediaType> accept = new ArrayList<>();
            ResteasyViolationException violationException = new ResteasyViolationException(container, accept);
            exceptionToUse.set(violationException);
            try {
                Thread.sleep(5);
            } catch (InterruptedException e) {
                //ignore
            }
            if (!run.get()) {
                break;
            }
        }

        //stop executors
        run.set(false);
        executor.shutdownNow();
        executor.awaitTermination(1, TimeUnit.MINUTES);


        //check if we had any exceptions in the runners
        boolean anyExceptions = false;
        for (Future<Boolean> result : runnerResults) {
            anyExceptions |= result.get();
        }
        assertFalse(anyExceptions, "Didn't expect any exceptions");
    }

    /**
     * This test works, because the toString method of the exception is set to exceptionToUse. This way only there are no
     * concurrency issues in the toString method. (Only one thread in org.jboss.resteasy.api.validation.ResteasyViolationException#convertViolations()
     * )
     */
    @Test
    void exceptionDoesntOccureIfToStringIsCalledInOneThreadFirst() throws Exception {
        //setup runners
        ExecutorService executor = Executors.newFixedThreadPool(10);
        AtomicBoolean run = new AtomicBoolean(true);
        AtomicReference<ResteasyViolationException> exceptionToUse = new AtomicReference<>();

        //start runners
        List<Future<Boolean>> runnerResults = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            runnerResults.add(executor.submit(new ExceptionToStringCaller(run, exceptionToUse)));
        }

        //start create exceptions
        for (int i = 0; i < 1_000; i++) {
            Set<ConstraintViolation<Object>> violations = new HashSet<>();
            violations.add(new DummyConstraintViolation());
            violations.add(new DummyConstraintViolation());
            violations.add(new DummyConstraintViolation());
            final SimpleViolationsContainer container = new SimpleViolationsContainer(violations);
            final List<MediaType> accept = new ArrayList<>();
            ResteasyViolationException violationException = new ResteasyViolationException(container, accept);

            //PREVENTS EXCEPTION
            violationException.toString();


            exceptionToUse.set(violationException);
            try {
                Thread.sleep(5);
            } catch (InterruptedException e) {
                //ignore
            }
            if (!run.get()) {
                break;
            }
        }

        //stop executors
        run.set(false);
        executor.shutdownNow();
        executor.awaitTermination(1, TimeUnit.MINUTES);


        //check if we had any exceptions in the runners
        boolean anyExceptions = false;
        for (Future<Boolean> result : runnerResults) {
            anyExceptions |= result.get();
        }
        assertFalse(anyExceptions, "Didn't expect any exceptions");
    }

    /**
     * This test works, because the constructor org.jboss.resteasy.api.validation.ResteasyViolationException#ResteasyViolationException(org.jboss.resteasy.plugins.providers.validation.ViolationsContainer,
     * java.util.List) calls org.jboss.resteasy.api.validation.ResteasyViolationException#convertToStrings(org.jboss.resteasy.plugins.providers.validation.ViolationsContainer).
     * This way the org.jboss.resteasy.api.validation.ResteasyViolationException#violationLists is initialized in the constructor
     * and therefore there are no concurrency issues in the toString method.
     */
    @Test
    void exceptionDoesntOccureIfOtherConstructorsAreUsed() throws Exception {
        //setup runners
        ExecutorService executor = Executors.newFixedThreadPool(10);
        AtomicBoolean run = new AtomicBoolean(true);
        AtomicReference<ResteasyViolationException> exceptionToUse = new AtomicReference<>();

        //start runners
        List<Future<Boolean>> runnerResults = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            runnerResults.add(executor.submit(new ExceptionToStringCaller(run, exceptionToUse)));
        }

        //Helper to reuse the convert-logic from ConstraintViolation to ResteasyConstraintViolation
        ConvertHelper helper = new ConvertHelper();

        //start create exceptions
        for (int i = 0; i < 1_000; i++) {
            //USE ViolationsContainer instead of SimpleViolationsContainer prevents exception
            Set<ResteasyConstraintViolation> violations = new HashSet<>();
            //convert ConstraintViolations to ResteasyConstraintViolation first
            violations.add(helper.convertViolation(new DummyConstraintViolation()));
            violations.add(helper.convertViolation(new DummyConstraintViolation()));
            violations.add(helper.convertViolation(new DummyConstraintViolation()));

            //Use ViolationContainer instead of SimpleViolationsContainer to use "working" constructor
            final ViolationsContainer container = new ViolationsContainer(violations);
            final List<MediaType> accept = new ArrayList<>();

            ResteasyViolationException violationException = new ResteasyViolationException(container, accept);
            exceptionToUse.set(violationException);
            try {
                Thread.sleep(5);
            } catch (InterruptedException e) {
                //ignore
            }
            if (!run.get()) {
                break;
            }
        }

        //stop executors
        run.set(false);
        executor.shutdownNow();
        executor.awaitTermination(1, TimeUnit.MINUTES);


        //check if we had any exceptions in the runners
        boolean anyExceptions = false;
        for (Future<Boolean> result : runnerResults) {
            anyExceptions |= result.get();
        }
        assertFalse(anyExceptions, "Didn't expect any exceptions");
    }


    static class ConvertHelper extends ResteasyViolationException {

        public ConvertHelper() {
            super(new ViolationsContainer<>(""));
        }

        //make the protected method public
        @Override
        public ResteasyConstraintViolation convertViolation(ConstraintViolation<?> violation) {
            return super.convertViolation(violation);
        }
    }

    static class ExceptionToStringCaller implements Callable<Boolean> {

        final AtomicReference<ResteasyViolationException> exceptionToUse;
        final AtomicBoolean run;

        public ExceptionToStringCaller(AtomicBoolean run,
                AtomicReference<ResteasyViolationException> exceptionToUse) {
            this.run = run;
            this.exceptionToUse = exceptionToUse;
        }

        @Override
        public Boolean call() throws Exception {
            while (run.get()) {
                try {
                    if (exceptionToUse.get() != null) {
                        //EXCEPTION OCCURS HERE
                        String s = exceptionToUse.get().toString();
                        if (s.isEmpty()) {
                            throw new RuntimeException("No to String value");
                        }
                    }
                } catch (ConcurrentModificationException e) {
                    e.printStackTrace();
                    run.set(false);
                    return true;
                }
                try {
                    Thread.sleep(1);
                } catch (InterruptedException e) {
                    //ignore
                }
            }
            return false;
        }
    }

    /**
     * Helperclass to create violations without any dependencies
     */
    static class DummyConstraintViolation implements ConstraintViolation<Object> {

        @Override
        public String getMessage() {
            return "message";
        }

        @Override
        public String getMessageTemplate() {
            return "";
        }

        @Override
        public String getRootBean() {
            return "Root";
        }

        @Override
        public Class getRootBeanClass() {
            return String.class;
        }

        @Override
        public Object getLeafBean() {
            return "leaf";
        }

        @Override
        public Object[] getExecutableParameters() {
            return new Object[0];
        }

        @Override
        public Object getExecutableReturnValue() {
            return null;
        }

        @Override
        public Path getPropertyPath() {
            return new Path() {
                @Override
                public Iterator<Node> iterator() {
                    Node node = new Node() {
                        @Override
                        public String getName() {
                            return "name";
                        }

                        @Override
                        public boolean isInIterable() {
                            return false;
                        }

                        @Override
                        public Integer getIndex() {
                            return 0;
                        }

                        @Override
                        public Object getKey() {
                            return "key";
                        }

                        @Override
                        public ElementKind getKind() {
                            return ElementKind.BEAN;
                        }

                        @Override
                        public <T extends Node> T as(Class<T> nodeType) {
                            return (T) this;
                        }
                    };
                    return Collections.singleton(node).iterator();
                }
            };
        }

        @Override
        public Object getInvalidValue() {
            return "invalid";
        }

        @Override
        public ConstraintDescriptor<?> getConstraintDescriptor() {
            return null;
        }

        @Override
        public Object unwrap(Class type) {
            return this;
        }
    }
}
